local _, fluffy = ...

local tooltip_weapon = CreateFrame('GameTooltip', "WeaponTooltip", UIParent, 'GameTooltipTemplate');
tooltip_weapon:SetOwner(WorldFrame, "ANCHOR_NONE");
tooltip_weapon:AddFontStrings(
    tooltip_weapon:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
    tooltip_weapon:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" ) 
);

local function get_weapon_speed(tooltip_line)
	local out = 0;

	local splt = mysplit_speed(tooltip_line);
	local val = tonumber(splt[2]);

	if val ~= nil then
		out = val;
	end
	wipe(splt);
	return out;
end

local function get_weapon_damage_range(tooltip_line)
	local dmg_min = 0;
	local dmg_max = 0;

	local splt = mysplit_damage(tooltip_line);

	local nsplits = #splt;
	local idx1 = 1;
	local idx2 = 2;

	if #splt >= 2 then
		local val1 = tonumber(splt[idx1]);
		local val2 = tonumber(splt[idx2]);
	
		if val1 ~= nil then
			dmg_min = val1;
			dmg_max = val1;
		end
	
		if val2 ~= nil then
			dmg_max = val2;
		end
	end


	wipe(splt);
	return dmg_min, dmg_max;
end

local function parse_weapon_info(slot_id, db)

	local item_id_ = GetInventoryItemID("player", slot_id);
	local dmg_min;
	local dmg_max;
	local speed;

	if item_id_ ~= nil and item_id_ ~= 0 then
		if db[item_id_] == nil then
			local itemName, itemLink = GetItemInfo(item_id_);
		
			tooltip_weapon:ClearLines()
			tooltip_weapon:SetOwner(WorldFrame, "ANCHOR_NONE");
			
			if pcall(function() tooltip_weapon:SetHyperlink(itemLink) end) then
				local rstr = WeaponTooltipTextRight1:GetText();
				local lstr = WeaponTooltipTextLeft1:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight2:GetText();
				lstr = WeaponTooltipTextLeft2:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight3:GetText();
				lstr = WeaponTooltipTextLeft3:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight4:GetText();
				lstr = WeaponTooltipTextLeft4:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight5:GetText();
				lstr = WeaponTooltipTextLeft5:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight6:GetText();
				lstr = WeaponTooltipTextLeft6:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight7:GetText();
				lstr = WeaponTooltipTextLeft7:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
	
				rstr = WeaponTooltipTextRight8:GetText();
				lstr = WeaponTooltipTextLeft8:GetText();
				if rstr ~= nil then
					dmg_min, dmg_max = get_weapon_damage_range(lstr);
					speed = get_weapon_speed(rstr);
				end
				-- print(dmg_min, dmg_max, speed);
				db[item_id_] = {dmg_min, dmg_max, speed};
			end

			tooltip_weapon:Hide();
		end
	end
end

local function update_ranged_weapon_stats()
    if fluffy.is_player_hunter == false then
		return;
	end    

	fluffy.ranged_weapon_id = 0;
	fluffy.ranged_dmg_min = 0;
	fluffy.ranged_dmg_max = 0;
	fluffy.ranged_base_speed = 0;
	fluffy.ranged_dmg_avg = 0;

	if FluffyDBPC == nil then
		FluffyDBPC = {};
	end

	if FluffyDBPC["ranged_weapons"] == nil then
		FluffyDBPC["ranged_weapons"] = {};
	end

	parse_weapon_info(18, FluffyDBPC["ranged_weapons"]);

	local item_string = GetInventoryItemLink("player",18);
	
	local i_ = 0;
	local item_id = 0;
	local enchant_id = 0;
	if item_string ~= nil then
		for str in string.gmatch(item_string, "([^:]+)") do
			if i_ == 1 then
				item_id = tonumber(str);
			elseif i_ == 2 then
				enchant_id = tonumber(str);
				break;
			end
			
			i_ = i_ + 1;
		end
	end

	if item_id == 0 then
		return;
	end
	
	local weapon_data = FluffyDBPC["ranged_weapons"][item_id];

	if weapon_data ~= nil then
		fluffy.ranged_dmg_min = weapon_data[1];
		fluffy.ranged_dmg_max = weapon_data[2];
		fluffy.ranged_base_speed = weapon_data[3];
		fluffy.ranged_weapon_id = item_id;
	elseif weapon_data[3]  <= 0.05 then
		weapon_data = nil;
		-- update_ranged_weapon_stats();
		return;
	end

	local bonus_dmg = 0;
	if enchant_id == 30 then
		bonus_dmg = 1; -- crude scope
	elseif enchant_id == 32 then
		bonus_dmg = 2; -- standard scope
	elseif enchant_id == 33 then
		bonus_dmg = 3; -- accurate scope
	elseif enchant_id == 663 then
		bonus_dmg = 5; -- deadly scope
	elseif enchant_id == 664 then
		bonus_dmg = 7; -- sniper scope
	elseif enchant_id == 2523 then
		fluffy.ranged_hit = 3;-- blitz scope
	end

	fluffy.ranged_dmg_min = fluffy.ranged_dmg_min + bonus_dmg;
	fluffy.ranged_dmg_max = fluffy.ranged_dmg_max + bonus_dmg;
	fluffy.ranged_dmg_avg = (fluffy.ranged_dmg_min + fluffy.ranged_dmg_max) / 2;

	-- if item_id ~= nil then
	-- 	print(FluffyDBPC["ranged_weapons"][item_id][1], FluffyDBPC["ranged_weapons"][item_id][2], FluffyDBPC["ranged_weapons"][item_id][3]);
	-- end
end

local function update_melee_weapon_stats()
    if fluffy.is_player_hunter == false then
		return;
	end    

	fluffy.melee_dmg_avg_main = 0;
	fluffy.main_hand_base_speed = 1;
	fluffy.melee_dmg_avg_off = 0;
	fluffy.off_hand_base_speed = 1;
	fluffy.melee_mh_weapon_id = 0;
	fluffy.melee_oh_weapon_id = 0;
		
	if FluffyDBPC == nil then
		FluffyDBPC = {};
	end

	if FluffyDBPC["melee_weapons"] == nil then
		FluffyDBPC["melee_weapons"] = {};
	end

	-- main hand
	parse_weapon_info(16, FluffyDBPC["melee_weapons"]);
	local item_id_mh = GetInventoryItemID("player", 16);

	-- off hand
	parse_weapon_info(17, FluffyDBPC["melee_weapons"]);
	local item_id_oh = GetInventoryItemID("player", 17);

	if item_id_mh ~= nil then
		local weapon_data = FluffyDBPC["melee_weapons"][item_id_mh];
		if weapon_data ~= nil then
			fluffy.melee_dmg_avg_main = 0.5 * (weapon_data[1] + weapon_data[2]);
			fluffy.main_hand_base_speed = weapon_data[3];
			fluffy.melee_mh_weapon_id = item_id_mh;
		elseif FluffyDBPC["melee_weapons"][item_id_mh][3]  <= 0.05 then
			FluffyDBPC["melee_weapons"][item_id_mh] = nil;
			-- update_melee_weapon_stats();
			return;
		end
	end

	if item_id_oh ~= nil then
		local weapon_data = FluffyDBPC["melee_weapons"][item_id_oh];
		if weapon_data ~= nil then
			fluffy.melee_dmg_avg_off = 0.5 * (weapon_data[1] + weapon_data[2]);
			fluffy.off_hand_base_speed = weapon_data[3];
			fluffy.melee_oh_weapon_id = item_id_oh;
		elseif FluffyDBPC["melee_weapons"][item_id_oh][3]  <= 0.05 then
			FluffyDBPC["melee_weapons"][item_id_oh] = nil;
			-- update_melee_weapon_stats();
			return;
		end
	end

	-- if item_id_mh ~= nil then
	-- 	print(FluffyDBPC["melee_weapons"][item_id_mh][1], FluffyDBPC["melee_weapons"][item_id_mh][2], FluffyDBPC["melee_weapons"][item_id_mh][3]);
	-- end
	-- if item_id_oh ~= nil then
	-- 	print(FluffyDBPC["melee_weapons"][item_id_oh][1], FluffyDBPC["melee_weapons"][item_id_oh][2], FluffyDBPC["melee_weapons"][item_id_oh][3]);
	-- end
end

function update_weapon_stats()
	update_ranged_weapon_stats();
	update_melee_weapon_stats();
end